/*******************************************************************************
*  EEPROM_MNT - EEPROMのメインテナンスツール                                   *
*                                                                              *
*    ArduinoIDEのCOM画面から操作してEEPROMの内容を変更／表示します。           *
*    IDE0022 & Arduino Duemilanove 328にて開発しています。                     *
*    このソフトはPDS(Public Domain Software)とします、自由に改造してＯＫです   *
*                                                                              *
* ============================================================================ *
*  VERSION DATE        BY                    CHANGE/COMMENT                    *
* ---------------------------------------------------------------------------- *
*  01.01   2011-03-26  きむ茶工房(きむしげ)  Create(初版公開)                  *
*  01.01   2011-03-29  きむ茶工房(きむしげ)  Melody(V1.0)機能を拡張            *
*  01.10   2011-04-04  きむ茶工房(きむしげ)  ファイル転送機能追加＆メモリ節約  *
*******************************************************************************/
#include <avr/pgmspace.h>
#include <EEPROM.h>
#include <ctype.h>
#include <string.h>

#define EXTENSION "(Melody V1.1)"  // 拡張は有り(Melody)タイプ
#define VER       "01.10"          // バージョン情報
#define EEPROMSIZE 1024            // EEPROM メモリーのサイズ

union {                            // 整数値とBYTEの変換用共用体の宣言
     char         cdt[2] ;
     unsigned int idt ;
} variable ;

int Address ;                      // 読書きするデータのEEPROM開始アドレス
int EndAdrs ;                      // このアドレスまでデータを送信します
                                   // 表示用メッセージデータ
prog_char Msg1[] PROGMEM = "Over EEPROM memory size!!\n" ;
prog_char Msg2[] PROGMEM = "Address Input Error!!\n" ;
prog_char Msg3[] PROGMEM = "Change to Address " ;
prog_char Msg4[] PROGMEM = "Input Error!!\n" ;
prog_char Msg5[] PROGMEM = "Command [R:Return Main] " ;
prog_char Msg6[] PROGMEM = "[S<Adrs(HEX)>:Set Adrs]" ;
prog_char Msg7[] PROGMEM = "Addresses can not return more!!\n" ;


#define MELODYSU   14              // メロディ用表示個数
int  MelodyAdrs ;                  // メロディ用セットするデータの開始アドレス
int  MelodyInputFlag ;             // メロディ用フラグ
char MelodyComd1[14][4] = {"do ","do#","re ","re#","mi ","mi#","fa ","fa#","so ","so#","ra ","ra#","si ","si#"} ;
char MelodyComd2[5][3] = {"en","ff","te","oc","qf"} ;
int  MelodyNote[17] = {0,1,2,4,8,16,32,64,128,1,2,4,8,16,32,64,128} ;
int  MelodyScale[5][14] = { {  65, 69, 73, 78, 82, 82, 87, 93, 98,104,110,117,123,123 },
                            { 131,139,147,156,165,165,175,185,196,208,220,233,247,247 },
                            { 262,277,294,311,330,330,349,370,392,415,440,466,494,494 },
                            { 523,554,587,622,659,659,698,740,784,831,880,932,988,988 },
              { 1047,1109,1175,1245,1319,1319,1397,1480,1568,1661,1760,1865,1976,1976 } } ;
prog_char MelodyMenu[] PROGMEM = " [I:Input] [D<t>:Ddisplay] [F]\n" ;
prog_char MelodyMsg1[] PROGMEM = "Octave(Range) ";
prog_char MelodyMsg2[] PROGMEM = "Temp = " ;
prog_char MelodyMsg3[] PROGMEM = "qf " ;


/*******************************************************************************
*  プログラムメモリ領域からデータを読み出してシリアル通信を行う                *
*******************************************************************************/
void Serial_print(const prog_char *s) {
  int i;
  for (i = 0; i < strlen_P(s); i ++) {
    Serial.write(pgm_read_byte(&s[i]));
  }
}
/*******************************************************************************
*  電源起動時とリセットの時だけのみ処理される関数(初期化と設定処理)            *
*******************************************************************************/
void setup() {
     Serial.begin(9600) ;          // シリアル通信の初期化
     Address = 0 ;
     EndAdrs = 0 ;
     MainMenuDsp() ;               // メインメニューの表示
}
/*******************************************************************************
*  繰り返し実行される処理の関数(メインの処理)                                  *
*******************************************************************************/
void loop() {
     char c ;     

     // ＩＤＥのＣＯＭからコマンド受信待ち
     if (Serial.available() > 0) {
          c = tolower(Serial.read()) ;
          switch (c) {
            case 't':              // EEPROMのメモリー内容転送処理
                     Transfer() ;
                     break ;
            case 'd':              // EEPROMのメモリー内容表示処理
                     Display() ;
                     break ;
            case 'e':              // EEPROMのメモリー内容変更処理
                     Edit() ;
                     break ;
            case 'a':              // このソフトについて
                     About() ;
                     break ;
            case 'm':              // 拡張機能(音楽メロディ作成)
                     Melody() ;
                     break ;
          }
          Serial.flush() ;
     }
}
// このスケッチについての情報を表示する
void About()
{
     Serial_print(PSTR("\nVersion ")) ;
     Serial.print(VER) ;
     Serial.println(EXTENSION) ;
     Serial_print(PSTR("This software maintenance ArduinIDE of the COM in the EEPROM.\n")) ;
     Serial_print(PSTR("The Arduino Duemilanove 328 programming aimed at.\n")) ;
     Serial_print(PSTR("This software is a PDS(Public Domain Software).\n")) ;
     Serial_print(PSTR("(Japanese site)http://www.geocities.jp/zattouka/GarageHouse/index.htm\n")) ;
     Serial_print(PSTR("by Kimucha Workshop(KimuShige)\n\n")) ;
}
/*******************************************************************************
*  MainMenuDsp - メインメニューの操作コマンドを表示する処理                    *
*******************************************************************************/
void MainMenuDsp()
{
     Serial_print(PSTR("\n<<< Arduino EEPROM Maintenance Tool Ver.")) ;
     Serial.print(VER) ;
     Serial_print(PSTR(" >>>\n")) ;
     Serial_print(PSTR("Menu Command [E:EEPROM Edit] [D:EEPROM display] [T:Transfer] [a:about this] [M:Melody]\n")) ;
}
/*******************************************************************************
*  AdrsDsp - 現在設定されているアドレスを表示する処理                          *
*******************************************************************************/
void AdrsDsp()
{
     char buf[6] ;
     
     Serial_print(PSTR("Now the Address of -> ")) ;
     sprintf(buf,"%04X",Address) ;
     Serial.println(buf) ;
}
/*******************************************************************************
*  AdrsSet - EEPROMに書込むアドレスを変更する処理                              *
*******************************************************************************/
void AdrsSet()
{
     int adrs ;

     // アドレスデータを受信する
     adrs = AdrsRCV() ;
     // 入力アドレスのチェック処理
     if (adrs >= EEPROMSIZE) {
          Serial_print(Msg2) ;          // アドレス入力エラー
     } else {
          Address = adrs ;
     }
}
// アドレス情報を受信して数値に変換する処理
int AdrsRCV()
{
     char buf[6] ;
     int adrs , i ;

     // 表示開始のアドレスを受信する(HEXで入力される物とする)
     memset(buf,0,sizeof(buf)) ;
     for (i=0 ; i<4 ; i++) {
          delay(10) ;
          buf[i] = Serial.read() ;
          if (Serial.available() == 0) break ;
     }
     Serial.flush() ;
     // 入力されたＨＥＸ文字アドレスを数値に変換する
     sscanf(buf,"%x",&adrs) ;

     return adrs ;
}
/*******************************************************************************
*                                                                              *
*  Melody - EEPROMのメモリーにメロディを作成する処理(オリジナル機能)           *
*                                                                              *
*******************************************************************************/
void Melody()
{
     char c ;
     int  flg , sw ;

     flg = 1 ;
     // メモリーにメロディを作成する操作のメニューコマンドを表示する
     Serial.write('\n') ;
     MelodyMenuDsp() ;
     MelodyInputFlag = 0 ;
     sw = 1 ;
     // ＩＤＥのＣＯＭからコマンド受信待ち
     do {
          if (Serial.available() > 0) {
               c = tolower(Serial.read()) ;
               switch (c) {
                 case 'r':              // メインメニューに戻る
                          flg = 0 ;
                          MelodyEnd() ; // データ入力終了印の0xFFを入力するか確認
                          MainMenuDsp() ;
                          break ;
                 case 's':              // 開始するアドレスを変更する
                          AdrsSet() ;
                          MelodyMenuDsp() ;
                          sw = 1 ;
                          break ;
                 case 'i':              // メロディのデータを入力する
                          if (sw == 0) AdrsDsp() ;
                          MelodyInput() ;
                          MelodyMenuDsp() ;
                          sw = 1 ;
                          break ;
                 case 'd':              // メロディのデータを表示する
                          if (sw == 1) MelodyAdrs = Address ;
                          MelodyDsp() ;
                          Serial_print(Msg5) ;
                          Serial_print(Msg6) ;
                          Serial_print(MelodyMenu) ;
                          sw = 0 ;
                          break ;
                 case 'f':              // データ入力終了印の0xFFを入力する
                          MelodyEndSub() ;
                          MelodyMenuDsp() ;
                          sw = 1 ;
                          break ;
               }
               Serial.flush() ;
          }
     } while(flg) ;
}
/*******************************************************************************
*  MelodyMenuDsp - メロディを作成する操作メニューコマンドを表示する処理        *
*******************************************************************************/
void MelodyMenuDsp()
{
     Serial_print(PSTR("<<< Create a Melody EEPROM >>>\n")) ;
     Serial_print(Msg5) ;
     Serial_print(Msg6) ;
     Serial_print(MelodyMenu) ;
     AdrsDsp() ;
}
/*******************************************************************************
*  MelodyEnd - メロディデータの入力終了時確認を行う処理                        *
*******************************************************************************/
void MelodyEnd()
{
     // メロディデータが入力されたか？
     if (MelodyInputFlag == 1) {
          MelodyEndSub() ;
     }
}
// メロディのデータ入力終了印の0xFFをセットする処理
void MelodyEndSub()
{
     char dt ;

     Serial.write('\n') ;
     AdrsDsp() ;                        // 現在の設定アドレスを表示
     Serial_print(PSTR("Want to enter complete 0xFF ? (Y/N) = ")) ;
     while(Serial.available() == 0) ;   // 入力されるまで待つ
     dt = Serial.read() ;
     Serial.flush() ;
     Serial.println(dt) ;               // (Y/N)を表示する
     // yesなら0xFFデータをセットする
     if ( (dt == 'y') || (dt == 'Y') ) {
          MelodyInputFlag = 0 ;
          EEPROM.write(Address,(char)0xff);
          Address++ ;
     }
}
/*******************************************************************************
*  MelodyDsp - メロディデータの表示を行う処理                                  *
*******************************************************************************/
void MelodyDsp()
{
     int  adrs ;
     char dt ;

     delay(10) ;
     // テンポ情報も表示する指示が有るか調べる
     if (Serial.available() != 0) {
          dt = Serial.read() ;
          Serial.flush() ;
          // テンポの表示を行う
          Serial_print(MelodyMsg2) ;
          Serial.print((unsigned char)EEPROM.read(MelodyAdrs),DEC) ;
          Serial_print(PSTR("M.M.\n")) ;
          MelodyAdrs++ ;
     }
     adrs = MelodyAdrs ;
     // アドレスの表示
     MelodyDspAdrs(adrs) ;
     // 音階の表示
     MelodyDspScale(adrs) ;
     // 音符の表示
     MelodyDspNote(adrs) ;
     // 音域の表示
     MelodyDspOctave(adrs) ;
     MelodyAdrs = MelodyAdrs + ((MELODYSU)*2) ;
}
/*******************************************************************************
*  MelodyDspAdrs - メロディデータのアドレス表示を行う処理                      *
*******************************************************************************/
void MelodyDspAdrs(int adrs)
{
     int i , cnt ;
     char buf[6] ;

     Serial_print(PSTR("Adrs  : ")) ;
     cnt = adrs + (MELODYSU * 2) ;
     for (i=adrs ; i<cnt ; i=i+2) {
          if (i >= EEPROMSIZE) break ;
          sprintf(buf,"%04X ",i) ;
          Serial.print(buf) ;
     }
     Serial.write('\n') ;
}
/*******************************************************************************
*  MelodyDspOctave - メロディデータの音域表示を行う処理                        *
*******************************************************************************/
void MelodyDspOctave(int adrs)
{
     unsigned int dt1 , dt2  ;
     int i , cnt , oc ;
     char buf[8] ;
          
     Serial_print(PSTR("Octave: ")) ;
     cnt = adrs + (MELODYSU * 2) ;
     for (i=adrs ; i<cnt ; i=i+2) {
          if (i >= EEPROMSIZE) break ;
          dt1 = EEPROM.read(i) ;
          if (dt1 == 0xff) break ;           // データが0xFFなら処理中断
          dt2 = (unsigned int)EEPROM.read(i+1) >> 5 ;
          dt1 = (dt1 << 3) | (dt2 & 0x7) ;   // 音階
          oc = MelodyScaleChg(dt1,0) ;
          if (oc < 7) sprintf(buf,"  %d  ",oc) ;
          else        strcpy(buf,"  ?  ") ;
          Serial.print(buf) ;
     }
     Serial.write('\n') ;
}
/*******************************************************************************
*  MelodyDspNote - メロディデータの音符表示を行う処理                          *
*******************************************************************************/
void MelodyDspNote(int adrs)
{
     unsigned int dt ;
     int i , cnt ;
     char buf[6] ;

     Serial_print(PSTR("Note  : ")) ;
     cnt = adrs + (MELODYSU * 2) ;
     for (i=adrs ; i<cnt ; i=i+2) {
          if (i >= EEPROMSIZE) break ;
          dt = EEPROM.read(i+1) ;
          if (dt == 0xff) break ;            // データが0xFFなら処理中断
          dt = dt & 0x1f ;                   // 音符データ
          if (dt >= 9) {
               if (dt <= 16) sprintf(buf,"%3d.",MelodyNote[dt]) ;
               else strcpy(buf,"    ") ;
          } else sprintf(buf,"%3d ",MelodyNote[dt]) ;
          Serial.print(buf) ;                // 音符表示
          Serial.write(' ') ;
     }
     Serial.write('\n') ;
}
/*******************************************************************************
*  MelodyDspScale - メロディデータの音階表示を行う処理                         *
*******************************************************************************/
void MelodyDspScale(int adrs)
{
     unsigned int dt1 , dt2 ;
     int i , cnt , oc ;
     char buf[6] ;

     Serial_print(PSTR("Scale :  ")) ;
     cnt = adrs + (MELODYSU * 2) ;
     for (i=adrs ; i<cnt ; i=i+2) {
          if (i >= EEPROMSIZE) break ;
          dt1 = EEPROM.read(i) ;
          dt2 = EEPROM.read(i+1) ;
          if (dt1 == 0xff) break ;           // データが0xFFなら処理中断
          dt2 = dt2 >> 5 ;
          dt1 = (dt1 << 3) | (dt2 & 0x7) ;   // 音階データ
          if (dt1 == 0) Serial_print(MelodyMsg3) ;
          else oc = MelodyScaleChg(dt1,1) ;  // 音階表示
          Serial.print("  ") ;
     }
     Serial.write('\n') ;
}
/*******************************************************************************
*  MelodyInput - メロディデータを入力する処理                                  *
*******************************************************************************/
void MelodyInput()
{
     char buf[8] , *p ;
     int  oc , note , scale;
     int  flg , i , x ;

     oc = 2 ;
     MelodyMsgDsp(oc) ;            // 入力操作コマンドを表示する
     flg = 1 ;
     do {
          // コマンド受信待ち
          memset(buf,0,sizeof(buf)) ;
          MelodyInputSub((char *)buf) ;
          // 終了のチェック
          if ((buf[0] == 'r')&&(buf[1]==0x00)) {
               strcpy(buf,"en") ;
          }
          // アドレスの移動処理
          if (buf[0] == '>') {
               MelodyAdrsNext() ;
               continue ;
          }
          if (buf[0] == '<') {
               MelodyAdrsPrev() ;
               continue ;
          }
          // 音符と音階の処理
          for (i=0 ; i<14 ; i++) {
               if (memcmp(buf,&MelodyComd1[i][0],2) == 0) break ;
          }
          if (i >= 0 && i<= 13) {
               p = (char *)memchr(buf,'.',6) ;
               if (p != 0) *p = 0x00 ;
               // 半音あげるかチェックする
               if (buf[2] == '#') {
                    sscanf(&buf[3],"%d",&note) ;
                    scale = MelodyScale[oc][i+1] ;
               } else {
                    sscanf(&buf[2],"%d",&note) ;
                    scale = MelodyScale[oc][i] ;
               }
               // 書き込み
               MelodyDataSet(scale,note,p) ;
               continue ;
          }
          // その他のコマンド処理
          for (i=0 ; i<4 ; i++) {
               if (memcmp(buf,&MelodyComd2[i][0],2) == 0) break ;
          }
          switch(i) {
            case  0:               // 入力終了
                    flg = 0 ;
                    break ;
            case  1:               // データ入力終了印の0xFFを入力する
                    MelodyEndSub() ;
                    MelodyMsgDsp(oc) ;
                    break ;
            case  2:               // テンポの入力を行う
                    sscanf(&buf[2],"%d",&x) ;
                    MelodyTempSet(x) ;
                    MelodyMsgDsp(oc) ;
                    break ;
            case  3:               // 音域(オクターブ)の設定処理
                    Serial_print(MelodyMsg1) ;
                    sscanf(&buf[2],"%d",&x) ;
                    if (x >= 2 && x <= 6) {
                         Serial.print(oc+2,DEC) ;
                         Serial_print(PSTR(" -> ")) ;
                         oc = x - 2 ;
                         Serial.println(oc+2,DEC) ;
                    } else {
                         Serial_print(Msg4) ;     // 入力エラー
                    }
                    break ;
            case  4:               // 休符の入力を行う
                    p = (char *)memchr(buf,'.',6) ;
                    if (p != 0) *p = 0x00 ;
                    sscanf(&buf[2],"%d",&note) ;
                    scale = 0 ;
                    // 書き込み
                    MelodyDataSet(scale,note,p) ;
                    break ;
          }
     } while(flg) ;
}
// 入力操作コマンドをＩＤＥから受信する
void MelodyInputSub(char *buf)
{
     int  i ;

     while(Serial.available() == 0) ;        // 入力されるまで待つ
     for (i=0 ; i<7 ; i++) {
          delay(10) ;
          *buf = tolower(Serial.read()) ;    // 大文字を小文字に変換
          if (Serial.available() == 0) break ;
          buf++ ;
     }
     Serial.flush() ;
}
/*******************************************************************************
*  MelodyMsgDsp - メモリー内容表示の入力操作コマンドを表示する処理             *
*******************************************************************************/
void MelodyMsgDsp(int oc)
{
     Serial_print(PSTR("[R:return] [TE:temp(10-250)] [OC:octave(2-6)] [FF] [<:prev] [>:next]\n")) ;
     Serial_print(PSTR("[do:re:mi:fa:so:ra:si:qf do<note(1-128)> (do4)(re#4)(ra8.)]\n")) ;
     Serial_print(MelodyMsg1) ;
     Serial_print(PSTR("-> ")) ;
     Serial.println(oc+2,DEC) ;    // 現在の音域(オクターブ)を表示する
}
/*******************************************************************************
*  MelodyAdrs??? - メロディデータの設定するアドレスを進めたり戻したりする処理  *
*******************************************************************************/
void MelodyAdrsNext()
{
     if ((EEPROMSIZE - Address) > 3) {
          Address = Address + 2 ;  // アドレスを進める
          MelodyAdrsSub() ;
     } else {
          Serial_print(Msg1) ;     // メモリサイズオーバー
          AdrsDsp() ;
     }
}
void MelodyAdrsPrev()
{
     if ((Address-2) > 0) {
          Address = Address - 2 ;  // アドレスを戻す
          MelodyAdrsSub() ;
     } else {
          Serial_print(Msg7) ;     // アドレスは戻せません
     }
}
// 現在アドレスのメロディデータ内容を表示する
void MelodyAdrsSub()
{
     unsigned int dt1 , dt2 ;
     int  oc ;
     char buf[16] ;

     // 現在アドレスの表示
     Serial_print(Msg3) ;
     sprintf(buf,": %04X ",Address) ;
     Serial.print(buf) ;
     dt1 = EEPROM.read(Address) ;
     if (dt1 == 0xff) {
          // データ終了表示
          Serial_print(PSTR("This data terminate (0xFF)\n")) ;
     } else {
          dt2 = (unsigned int)EEPROM.read(Address+1) >> 5 ;
          dt1 = (dt1 << 3) | (dt2 & 0x7) ;
          // 入力した音階の表示
          sprintf(buf,"[Scale:%4dHz ",dt1) ;
          Serial.print(buf) ;
          if (dt1 == 0) Serial_print(MelodyMsg3) ;
          else oc = MelodyScaleChg(dt1,1) ;
          // 入力した音符の表示
          dt1 = (unsigned int)EEPROM.read(Address+1) & 0x1f ;
          if (dt1 >= 9) {
               if (dt1 <= 16) sprintf(buf,"][Note:%3d.]",MelodyNote[dt1]) ;
               else strcpy(buf,"][Note:???]") ;
          } else sprintf(buf,"][Note:%3d]",MelodyNote[dt1]) ;
          Serial.print(buf) ;
          // 音域の表示
          Serial_print(PSTR("[Octave: ")) ;
          if (oc == 7) Serial.write('?') ;
          else         Serial.print(oc,DEC) ;
          Serial.println("]") ;
     }
}
/*******************************************************************************
*  MelodyScaleChg - 音階の周波数値からdo,re,mi,の文字を表示する処理            *
*    scale :音階の周波数値                                                     *
*    sw    :1で表示する　0:で表示しない                                        *
*    return:音域（オクターブ）を返す                                           *
*******************************************************************************/
int MelodyScaleChg(int scale,int sw)
{
     int i , j ;

     // 周波数値を検索する処理
     for (i=0 ; i<5 ; i++) {
          for (j=0 ; j<14 ; j++) {
               if (scale == MelodyScale[i][j]) break ;
          }
          if (j < 14) break ;
     }
     // ドレミの表示を行う処理
     if (sw == 1) {
          if (j < 14) Serial.print(&MelodyComd1[j][0]) ;
          else        Serial_print(PSTR("???")) ;  // 周波数値が無かった場合の表示
     }
     return (i+2) ;
}
/*******************************************************************************
*  MelodyTempSet - テンポのデータ書き込みを行う処理                            *
*    dt:設定するテンポの数値データ                                             *
*******************************************************************************/
void MelodyTempSet(int dt)
{
     // 書き込めるか調べる
     if ((EEPROMSIZE - Address) >= 2) {
          if (dt >= 10 && dt <= 250) {
               // 入力したテンポの表示を行う
               Serial_print(MelodyMsg2) ;
               Serial.print(dt,DEC) ;
               Serial_print(PSTR("M.M.\n")) ;
               // EEPROMにデータを書込む
               EEPROM.write(Address,(char)dt);
               Address++ ;
          } else {
               Serial_print(PSTR("Tempo ")) ;
               Serial_print(Msg4) ;          // 入力エラー
          }
     } else {
          Serial_print(Msg1) ;               // メモリサイズオーバー
          AdrsDsp() ;
     }
}
/*******************************************************************************
*  MelodyDataSet - メロディデータの書き込みを行う処理                          *
*    scale :音階の周波数値                                                     *
*    note  :音符の数値                                                         *
*    sw    :0で普通音符  0以外で付点音符                                       *
*******************************************************************************/
void MelodyDataSet(int scale,int note,char *sw)
{
     unsigned int dt , x ;
     int  oc ;
     char buf[16] ;

     // 書き込めるか調べる
     if ((EEPROMSIZE - Address) >= 2) {
          // 音符の数値をチェックする
          for (x=1 ; x<=8 ; x++) {
               if (MelodyNote[x] == note) break ;
          }
          if (x > 8) {
               Serial_print(PSTR("Note ")) ;
               Serial_print(Msg4) ;           // 入力エラー
          } else {
               // 現在アドレスの表示
               sprintf(buf,"[Adrs:%04X]",Address) ;
               Serial.print(buf) ;
               // 入力した音階と音符の表示
               sprintf(buf,"[Scale:%4dHz ",scale) ;
               Serial.print(buf) ;
               if (scale == 0) {
                    Serial_print(MelodyMsg3) ;
                    oc = 0 ;
               } else oc = MelodyScaleChg(scale,1) ;
               if (sw == 0) sprintf(buf,"][Note:%2d]",note) ;
               else {
                    x = x + 8 ;
                    sprintf(buf,"][Note:%2d.]",note) ;
               }
               Serial.print(buf) ;
               // 音域の表示
               sprintf(buf,"[Octave:%2d]",oc) ;
               Serial.println(buf) ;
               // EEPROMに音階の書き込み処理
               dt = (unsigned int)scale >> 3 ;// frequencyの下位３ビットを消す
               EEPROM.write(Address,(char)dt);// １バイト目書込み
               Address++ ;
               // EEPROMに音符の書き込み処理
               dt = (unsigned int)scale & 0x7 ;
               dt = (dt << 5) | (x & 0x1f) ;
               EEPROM.write(Address,(char)dt);// ２バイト目書込み
               Address++ ;
          }
     } else {
          Serial_print(Msg1) ;                // メモリサイズオーバー
          AdrsDsp() ;
     }
}
/*******************************************************************************
*                                                                              *
*  Transfer - EEPROMのメモリー内容を転送する処理                               *
*                                                                              *
*******************************************************************************/
void Transfer()
{
     char c , buf[6] ;
     int  adrs ;

     while(Serial.available() == 0) ;   // 送信されるまで待つ
     c = tolower(Serial.read()) ;
     Serial.flush() ;
     switch (c) {
       case 's':                        // 読み書きするEEPROMのアドレスを設定する
                Serial_print(PSTR("[arduino:]")) ;
                AdrsSet() ;
                AdrsDsp() ;
                break ;
       case 'w':                        // ＰＣから送信されるので受信してEEPROMに書込む
                Serial_print(PSTR("[arduino:Listen from the PC]\n")) ;
                TransferReceive() ;
                MainMenuDsp() ;         // メインメニューの表示
                break ;
       case 'r':                        // EEPROMから読込んでＰＣへ送信する
                Serial_print(PSTR("[arduino:Transmitted to the PC]\n")) ;
                TransferSend() ;
                MainMenuDsp() ;         // メインメニューの表示
                break ;
       case 'e':                        // ここで指定されたアドレスまでを送信する
                Serial_print(PSTR("[arduino:]")) ;
                adrs = AdrsRCV() ;
                // 入力アドレスのチェック処理
                if (adrs >= EEPROMSIZE) {
                     Serial_print(Msg2);// アドレス入力エラー
                } else {
                     EndAdrs = adrs ;
                     Serial_print(PSTR("That address the -> ")) ;
                     sprintf(buf,"%04X",EndAdrs) ;
                     Serial.println(buf) ;
                }
                break ;
     }
}
/*******************************************************************************
*  TransferReceive - ＰＣから送信されるので受信してEEPROMに書込む処理          *
*******************************************************************************/
void TransferReceive()
{
     char buf[6] ;
     int  adrs ;
     unsigned int  i , dt ;

     adrs = Address ;
     while(1) {
          while(Serial.available() == 0) ;   // 送信されるまで待つ
          memset(buf,0,sizeof(buf)) ;
          for (i=0 ; i<2 ; i++) {
               //delay(10) ;
               buf[i] = Serial.read() ;      // ２文字(HEX文字)受信する
               if (Serial.available() == 0) break ;
               if (buf[i] == 'X') break ;    // 転送終了
          }
          if (i == 2) {
               // 受信した１バイトＨＥＸ文字を数値に変換する
               sscanf(buf,"%x",&dt) ;
               // EEPROMにデータを書込む
               EEPROM.write(Address,(char)dt);
               Address++ ;
          } else {
               Serial.print(Address-adrs,DEC) ;
               Serial_print(PSTR(" bytes from the address written ")) ;
               sprintf(buf,"%04X",adrs) ;
               Serial.println(buf) ;
               Serial_print(PSTR("[arduino:EEPROM write end]\n")) ;
               break ;                       // 処理終了
          }
     }
     Serial.flush() ;
}
/*******************************************************************************
*  TransferSend - EEPROMから読込んでＰＣへ送信する処理                         *
*******************************************************************************/
void TransferSend()
{
     char buf[6] ;
     int  i ;

     for (i=Address ; i<=EndAdrs ; i++) {
          // EEPROMから読込みＨＥＸ文字に変換する
          sprintf(buf,"%02X",EEPROM.read(i)) ;
          // 送信する
          Serial.println(buf) ;
          delay(20) ;
     }
     Serial.println("XX") ;   // 送信完了
     delay(10) ;
     Serial.print((EndAdrs-Address)+1,DEC) ;
     Serial_print(PSTR(" bytes sent from the address ")) ;
     sprintf(buf,"%04X",Address) ;
     Serial.println(buf) ;
     Serial_print(PSTR("[arduino:transmission is complete]\n")) ;
}
/*******************************************************************************
*                                                                              *
*  Display - EEPROMのメモリー内容を表示する処理                                *
*                                                                              *
*******************************************************************************/
void Display()
{
     char c ;
     int  flg ;

     flg = 1 ;
     // メモリー内容の表示を行う操作のコマンドを表示する
     DisplayMenuDsp() ;
     // ＩＤＥのＣＯＭからコマンド受信待ち
     do {
          if (Serial.available() > 0) {
               c = tolower(Serial.read()) ;
               switch (c) {
                 case 'r':         // メインメニューに戻る
                          flg = 0 ;
                          break ;
                 case 'a':         // メモリー内容を全て表示する
                          Serial_print(PSTR("** EEPROM MEMORY All DUMP **\n")) ;
                          MemoryDsp(0,EEPROMSIZE) ;
                          break ;
                 case 'l':         // 指定アドレスから１６バイト表示する
                          DisplayLine() ;
                          break ;
               }
               Serial.flush() ;
               if (flg == 1) DisplayMenuDsp() ;
               else          MainMenuDsp() ;
          }
     } while(flg) ;
}
/*******************************************************************************
*  DisplayMenuDsp - メモリー内容表示の操作コマンドを表示する処理               *
*******************************************************************************/
void DisplayMenuDsp()
{
     Serial_print(PSTR("<<< EEPROM display >>>\n")) ;
     Serial_print(Msg5) ;
     Serial_print(PSTR("[A:ALL List] [L<Adrs(HEX)>:1Line List]\n")) ;
}
/*******************************************************************************
*  DisplayLine - 指定アドレスから１６バイト(１行分)表示する処理                *
*******************************************************************************/
void DisplayLine()
{
     char buf[6] ;
     unsigned int adrs , i ;

     // 表示開始のアドレスを受信する(HEXで入力される物とする)
     memset(buf,0,sizeof(buf)) ;
     for (i=0 ; i<5 ; i++) {
          delay(10) ;
          buf[i] = Serial.read() ;
          if (Serial.available() == 0) break ;
     }
     Serial.flush() ;
     // 入力されたＨＥＸ文字アドレスを数値に変換する
     sscanf(buf,"%x",&adrs) ;
     // 入力アドレスのチェック処理
     if (adrs >= EEPROMSIZE) Serial_print(Msg2) ;      // アドレス入力エラー
     else                    MemoryDsp(adrs,adrs+16) ; // 実際の表示を行う処理
}
/*******************************************************************************
*  MemoryDsp - メモリー内容を表示する処理                                      *
*    adrs   :表示を開始するアドレスを指定                                      *
*    memsize:表示するメモリのサイズを指定                                      *
*******************************************************************************/
void MemoryDsp(int adrs,int memsize)
{
     char buf[8] , c ;
     int i , j , k ;
     
     Serial_print(PSTR("Adrs : 00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F : 0123456789ABCDEF\n")) ;
     // EEPROMの指定アドレスから指定サイズだけ処理する
     for (i=adrs ; i < memsize ; i=i+16) {
          // メモリ１６バイト読み出す
          k=i+16 ;
          sprintf(buf,"%04X : ",i) ;
          Serial.print(buf) ;                          // アドレスの表示
          // １６バイトをＨＥＸで表示する
          for (j=i ; j<k ; j++) {
               if (j >= EEPROMSIZE) {
                    Serial.print("   ") ;
               } else {
                    sprintf(buf,"%02X ",EEPROM.read(j)) ;
                    Serial.print(buf) ;                // 内容をＨＥＸで表示
               }
          }
          Serial.print(": ") ;
          // １６バイトをアスキーで表示する
          for (j=i ; j<k ; j++) {
               if (j >= EEPROMSIZE) break ;
               c = EEPROM.read(j) ;
               if ((c >= 0x20) && (c <= 0x7e)) {
                    Serial.write(c) ;                  // 内容をアスキー文字で表示
               } else {
                    Serial.write('.') ;                // アスキーで表示できないデータは(.)で表示
               }
          }
          Serial.write('\n') ;
     }
}
/*******************************************************************************
*                                                                              *
*  Edit - EEPROMのメモリー内容を変更する処理                                   *
*                                                                              *
*******************************************************************************/
void Edit()
{
     char c ;
     int  flg ;

     flg = 1 ;
     // メモリー内容の変更を行う操作のコマンドを表示する
     EditMenuDsp() ;
     // ＩＤＥのＣＯＭからコマンド受信待ち
     do {
          if (Serial.available() > 0) {
               c = tolower(Serial.read()) ;
               switch (c) {
                 case 'r':                   // メインメニューに戻る
                          flg = 0 ;
                          MainMenuDsp() ;
                          break ;
                 case 's':                   // 設定するアドレスを変更する
                          AdrsSet() ;
                          EditMenuDsp() ;
                          break ;
                 case '>':EditAdrsNext() ;   // アドレスを１バイト進める
                          break ;
                 case '<':EditAdrsPrev() ;   // アドレスを１バイト戻す
                          break ;
                 case 'm':                   // 変更するデータを文字で入力する
                          EditMojiSet() ;
                          break ;
                 case 'i':                   // 変更するデータを整数値で入力する
                          EditIntSet() ;
                          break ;            // 変更するデータをＨＥＸで入力する
                 default :if ( (c >= '0' && c <= '9') ||
                               (c >= 'a' && c <= 'f') ) {
                               EditByteSet(c) ;
                          }
               }
               Serial.flush() ;
          }
     } while(flg) ;
}
/*******************************************************************************
*  EditMenuDsp - メモリー内容変更の操作コマンドを表示する処理                  *
*******************************************************************************/
void EditMenuDsp()
{
     char buf[6] ;

     Serial_print(PSTR("<<< EEPROM Edit >>>\n")) ;
     Serial_print(Msg5) ;
     Serial_print(Msg6) ;
     Serial_print(PSTR(" [<:Prev] [>:Next]\n")) ;
     Serial_print(PSTR("        [HEX:Byte Set] [I<Dec>:Int Set] [M<string>:Moji Set]\n")) ;
     AdrsDsp() ;
}
/*******************************************************************************
*  EditAdrs??? - データ設定を行うアドレスを１バイト進めたり戻したりする処理    *
*******************************************************************************/
void EditAdrsNext()
{
     if ((EEPROMSIZE - Address) <= 1) {
          Serial_print(Msg1) ;          // メモリサイズオーバー
     } else {
          Address++ ;
          EditAdrsSub() ;
     }

}
void EditAdrsPrev()
{
     if (Address <= 0) {
          Serial_print(Msg7) ;          // アドレスは戻せません
     } else {
          Address-- ;
          EditAdrsSub() ;
     }
}
// アドレスの表示とデータ内容をＨＥＸで表示する
void EditAdrsSub()
{
     char buf[14] ;

     sprintf(buf,"Adrs %04X : ",Address) ;
     Serial.print(buf) ;                 // 現在アドレスの表示
     Serial.println((unsigned char)EEPROM.read(Address),HEX) ;
}
/*******************************************************************************
*  EditByteSet - EEPROMに変更するデータをＨＥＸで１バイト書込む処理            *
*    c : 入力された上位４ビットのＨＥＸデータ                                  *
*******************************************************************************/
void EditByteSet(char c)
{
     char buf[6] ;
     int  dt ;

     memset(buf,0,sizeof(buf)) ;
     buf[0] = c ;                       // 上位４ビットのＨＥＸデータを設定
     delay(10) ;
     // 下位４ビットのＨＥＸデータを受信する
     buf[1] = tolower(Serial.read()) ;
     Serial.flush() ;
     if ( (buf[1] >= '0' && buf[1] <= '9') ||
          (buf[1] >= 'a' && buf[1] <= 'f') ) {
          if (Address >= EEPROMSIZE) {
               Serial_print(Msg1) ;     // メモリサイズオーバー
          } else {
               // 入力された１バイトＨＥＸ文字を数値に変換する
               sscanf(buf,"%x",&dt) ;
               // 変更情報の表示
               Serial_print(Msg3) ;     // アドレスに変更を行う
               sprintf(buf,"%04X:",Address) ;
               Serial.print(buf) ;
               Serial.print((unsigned char)EEPROM.read(Address),HEX) ;
               Serial.print(" -> ") ;
               Serial.println(dt,HEX) ;
               // EEPROMにデータを書込む
               EEPROM.write(Address,(char)dt);
               Address++ ;
          }
      }
}
/*******************************************************************************
*  EditMojiSet - EEPROMに変更するデータを文字で書込む処理                      *
*******************************************************************************/
void EditMojiSet()
{
     char buf[6] ;
     unsigned char c ;

     memset(buf,0,sizeof(buf)) ;
     // 入力された文字列だけ処理する
     do {
          delay(10) ;
          c = Serial.read() ;
          if ( (c >= 0x20 && c <= 0x7e) ||
               (c >= 0xa0 && c <= 0xdf) ) {
               if (Address >= EEPROMSIZE) {
                    Serial_print(Msg1) ;     // メモリサイズオーバー
                    break ;
               } else {
                    // 変更情報の表示
                    Serial_print(Msg3) ;     // アドレスに変更を行う
                    sprintf(buf,"%04X:",Address) ;
                    Serial.print(buf) ;
                    Serial.print((unsigned char)EEPROM.read(Address),HEX) ;
                    Serial.print(" -> ") ;
                    Serial.println((unsigned char)c,HEX) ;
                    // EEPROMにデータを書込む
                    EEPROM.write(Address,c);
                    Address++ ;
               }
          } else {
               Serial_print(PSTR("String ")) ;
               Serial_print(Msg4) ;          // 入力エラー
               break ;
          }
          if (Serial.available() == 0) break ;
     } while(1) ;
     Serial.flush() ;
}
/*******************************************************************************
*  EditIntSet - EEPROMに変更するデータを整数値で書込む処理                     *
*******************************************************************************/
void EditIntSet()
{
     char buf[8] ;
     int dt , i , sw ;

     // 変更するデータを受信する
     sw = 0 ;
     memset(buf,0,sizeof(buf)) ;
     for (i=0 ; i<6 ; i++) {
          delay(10) ;
          buf[i] = Serial.read() ;
          // 数字のデータかチェックする
          if (buf[i] < 0x30 || buf[i] > 0x39) {
               if (buf[i] != 0x2d) {
                    sw = 1 ;
                    break ;
               }
          }
          if (Serial.available() == 0) break ;
     }
     Serial.flush() ;
     if (sw == 0) {
          if ((EEPROMSIZE - Address) >= 2){
               // 入力された整数値の文字を数値に変換する
               sscanf(buf,"%d",&dt) ;
               // 変更情報の表示
               Serial_print(Msg3) ;     // アドレスに変更を行う
               sprintf(buf,"%04X:",Address) ;
               Serial.print(buf) ;
               Serial.print((int)uint_get(Address),DEC) ;
               Serial.print(" -> ") ;
               Serial.println(dt,DEC) ;
               // EEPRMにデータを書込む
               uint_set(Address,(unsigned int)dt) ;
               Address = Address + 2 ;
          } else {
               Serial_print(Msg1) ;     // メモリサイズオーバー
          }
     } else {
          Serial_print(PSTR("Int ")) ;
          Serial_print(Msg4) ;          // 入力エラー
     }
}
/*******************************************************************************
*  uint_set - EEPROMに整数値で書込む処理                                       *
*    add : EEPROMの書込むアドレスを指定                                        *
*    dt  : 書込む整数値のデータを指定                                          *
*******************************************************************************/
void uint_set(int add,unsigned int dt)
{
     variable.idt = dt ;
     EEPROM.write(add, variable.cdt[0]);
     EEPROM.write(add+1, variable.cdt[1]);
}
/*******************************************************************************
*  uint_get - EEPROMから整数値で読込む処理                                     *
*    add : EEPROMの読込むアドレスを指定                                        *
*    戻り: 読込まれた整数値を返します                                          *
*******************************************************************************/
unsigned int uint_get(int add)
{
     variable.cdt[0] = EEPROM.read(add) ;
     variable.cdt[1] = EEPROM.read(add+1) ;
     return variable.idt ;     
}
